/*
 * uci.c - helper functions for UCI config file.
 *
 * Copyright (C) 2024 Iopsys Software Solutions AB. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <limits.h>
#include <net/if.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <easy/easy.h>
#include <wifi.h>
#include <wifidefs.h>

#include "wifimngr.h"

static enum wifi_band uci_to_radio_band(const char *uci_band)
{
	if (!uci_band)
		return BAND_UNKNOWN;

	if (!strcmp(uci_band, "2g"))
		return BAND_2;
	else if (!strcmp(uci_band, "5g"))
		return BAND_5;
	else if (!strcmp(uci_band, "6g"))
		return BAND_6;
	else
		return BAND_UNKNOWN;
}

static void uci_add_option(struct uci_context *ctx, struct uci_package *p,
			   struct uci_section *s, const char *option,
			   void *value, bool is_list)
{
	struct uci_ptr ptr = { 0 };

	ptr.p = p;
	ptr.s = s;
	ptr.package = p->e.name;
	ptr.section = s->e.name;
	ptr.option = option;
	ptr.target = UCI_TYPE_OPTION;
	ptr.flags |= UCI_LOOKUP_EXTENDED;
	ptr.value = (char *)value;

	if (is_list)
		uci_add_list(ctx, &ptr);
	else
		uci_set(ctx, &ptr);
}

int uci_add_wifi_iface(char **argv)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_section *sec;
	struct uci_element *e;
	struct uci_option *op;
	char *ifname = NULL;
	bool found = false;
	int i = 0;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	while (argv[i]) {
		if (!strcmp(argv[i], "ifname")) {
			ifname = argv[i + 1];
			break;
		}
		i += 2;
	}

	if (ifname) {
		uci_foreach_element(&pkg->sections, e) {
			struct uci_element *x, *tmp;

			sec = uci_to_section(e);

			if (strcmp(sec->type, "wifi-iface"))
				continue;

			uci_foreach_element_safe(&sec->options, tmp, x) {
				op = uci_to_option(x);
				if (op->type == UCI_TYPE_STRING &&
				    !strncmp(x->name, "ifname", 6) &&
				    !strncmp(op->v.string, ifname, 15)) {

					found = true;
					break;
				}
			}

			if (found) {
				wifimngr_warn("%s %s already added\n", __func__, ifname);
				break;
			}
		}
	}

	if (!found)
		uci_add_section(ctx, pkg, "wifi-iface", &sec);

	i = 0;
	while (argv[i]) {
		uci_add_option(ctx, pkg, sec, argv[i], argv[i + 1], false);
		i += 2;
	}

	uci_commit(ctx, &pkg, false);
	uci_free_context(ctx);

	return 0;
}

int uci_del_wifi_iface(char *ifname)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	struct uci_option *op;
	bool found = false;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);
		struct uci_element *x, *tmp;

		if (strcmp(s->type, "wifi-iface"))
			continue;

		uci_foreach_element_safe(&s->options, tmp, x) {
			op = uci_to_option(x);
			if (op->type == UCI_TYPE_STRING &&
				!strncmp(x->name, "ifname", 6) &&
				!strncmp(op->v.string, ifname, 15)) {

				found = true;
				break;
			}
		}

		if (found) {
			struct uci_ptr ptr = { 0 };

			ptr.p = pkg;
			ptr.s = s;
			ptr.package = pkg->e.name;
			ptr.section = s->e.name;
			ptr.option = NULL;
			ptr.target = UCI_TYPE_SECTION;
			ptr.flags |= UCI_LOOKUP_EXTENDED;
			ptr.value = "wifi-iface";

			uci_delete(ctx, &ptr);
			uci_commit(ctx, &pkg, false);
			break;
		}
	}

	uci_free_context(ctx);

	return found ? 0 : -1;
}

int uci_get_wifi_devices(struct wifimngr *w, const char *conffile)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, conffile, &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	/* flush older wifimngr_device */
	memset(w->wdev, 0, WIFI_DEV_MAX_NUM * sizeof(struct wifimngr_device));
	w->num_wifi_device = 0;

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);
		if (!strcmp(s->type, "wifi-device")) {
			struct uci_element *x;
			struct uci_option *op;
			//int idx;


			//FIXME
			//idx = wifimngr_lookup_wifi_device(w, s->e.name);

			strncpy(w->wdev[w->num_wifi_device].device, s->e.name, 15);
			strncpy(w->wdev[w->num_wifi_device].phy, s->e.name, 15);

			if (!phy_dir_exist(s->e.name)) {
				uci_foreach_element(&s->options, x) {
					op = uci_to_option(x);
					if (!strncmp(x->name, "path", 4) &&
					    op->type == UCI_TYPE_STRING) {
						char phy[16] = {0};
						int ret;

						ret = find_phy_from_device_path(op->v.string, phy);
						if (!ret) {
							memset(w->wdev[w->num_wifi_device].phy, 0, 15);
							strncpy(w->wdev[w->num_wifi_device].phy, phy, 15);
						}
					}
				}
			}

			/* Get band */
			uci_foreach_element(&s->options, x) {
				op = uci_to_option(x);
				if (!strncmp(x->name, "band", 4) &&
				    op->type == UCI_TYPE_STRING) {
					w->wdev[w->num_wifi_device].band = uci_to_radio_band(op->v.string);
				}
			}

			/* Get disabled */
			uci_foreach_element(&s->options, x) {
				op = uci_to_option(x);
				if (!strncmp(x->name, "disabled", 8) &&
				    op->type == UCI_TYPE_STRING) {
					w->wdev[w->num_wifi_device].disabled = atoi(op->v.string);
				}
			}

			w->num_wifi_device++;
			if (w->num_wifi_device == WIFI_DEV_MAX_NUM)
				break;
		}
	}

	uci_free_context(ctx);
	return w->num_wifi_device;
}

int uci_get_wifi_interfaces(struct wifimngr *w, const char *conffile)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, conffile, &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	memset(w->ifs, 0, sizeof(struct wifimngr_iface) * WIFI_IF_MAX_NUM);
	w->num_wifi_iface = 0;

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "wifi-iface")) {
			struct uci_element *x;
			struct uci_option *op;

			uci_foreach_element(&s->options, x) {
				op = uci_to_option(x);
				if (!strncmp(x->name, "ifname", 6) &&
					op->type == UCI_TYPE_STRING) {

					strncpy(w->ifs[w->num_wifi_iface].iface, op->v.string, 15);
				} else if (!strncmp(x->name, "device", 6) &&
					op->type == UCI_TYPE_STRING) {

					strncpy(w->ifs[w->num_wifi_iface].device, op->v.string, 15);
				} else if (!strncmp(x->name, "disabled", 8) &&
					   op->type == UCI_TYPE_STRING) {

					w->ifs[w->num_wifi_iface].disabled = !!atoi(op->v.string);
				} else if (!strncmp(x->name, "mode", 4) &&
					op->type == UCI_TYPE_STRING) {

					if (!strncmp(op->v.string, "ap", 2))
						w->ifs[w->num_wifi_iface].mode = WIFI_AP;
					else if (!strncmp(op->v.string, "wet", 3))
						w->ifs[w->num_wifi_iface].mode = WIFI_STA;
					else if (!strncmp(op->v.string, "sta", 3))
						w->ifs[w->num_wifi_iface].mode = WIFI_STA;
				}
			}

			if (strlen(w->ifs[w->num_wifi_iface].iface) == 0)
				snprintf(w->ifs[w->num_wifi_iface].iface, 16, "idx%d", w->num_wifi_iface);

			w->num_wifi_iface++;
			if (w->num_wifi_iface == WIFI_IF_MAX_NUM)
				break;
		}
	}

	uci_free_context(ctx);
	return w->num_wifi_iface;
}

const char *uci_get_wifi_option(char *type, char *name, const char *option)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	const char *val = NULL;

	ctx = uci_alloc_context();
	if (!ctx)
		return NULL;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return NULL;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);
		if (!strcmp(s->type, type)) {
			val = uci_lookup_option_string(ctx, s, option);
			break;
		}
	}

	uci_free_context(ctx);
	return val ? val : NULL;
}

